@@ -18,10 +18,8 @@ import copy
 
 from oslo_config import cfg
 from oslo_log import log as logging
-import six
 
 from nova import exception
-from nova.i18n import _LE
 from nova.objects import fields
 from nova.objects import pci_device_pool
 from nova.pci import utils
@@ -65,7 +63,7 @@ class PciDeviceStats(object):
                       for pci_pool in stats] if stats else []
         self.pools.sort(key=lambda item: len(item))
         self.dev_filter = dev_filter or whitelist.Whitelist(
-            CONF.pci_passthrough_whitelist)
+            CONF.pci.passthrough_whitelist)
 
     def _equal_properties(self, dev, entry, matching_keys):
         return all(dev.get(prop) == entry.get(prop)
@@ -96,6 +94,14 @@ class PciDeviceStats(object):
         pool = {k: getattr(dev, k) for k in self.pool_keys}
         if tags:
             pool.update(tags)
+        # NOTE(gibi): parent_ifname acts like a tag during pci claim but
+        # not provided as part of the whitelist spec as it is auto detected
+        # by the virt driver.
+        # This key is used for match InstancePciRequest backed by neutron ports
+        # that has resource_request and therefore that has resource allocation
+        # already in placement.
+        if dev.extra_info.get('parent_ifname'):
+            pool['parent_ifname'] = dev.extra_info['parent_ifname']
         return pool
 
     def add_device(self, dev):
@@ -152,17 +158,21 @@ class PciDeviceStats(object):
             # a spec may be able to match multiple pools.
             pools = self._filter_pools_for_spec(self.pools, spec)
             if numa_cells:
-                pools = self._filter_pools_for_numa_cells(pools, numa_cells)
-            pools = self._filter_non_requested_pfs(request, pools)
+                numa_policy = None
+                if 'numa_policy' in request:
+                    numa_policy = request.numa_policy
+                pools = self._filter_pools_for_numa_cells(
+                    pools, numa_cells, numa_policy, count)
+            pools = self._filter_non_requested_pfs(pools, request)
             # Failed to allocate the required number of devices
             # Return the devices already allocated back to their pools
             if sum([pool['count'] for pool in pools]) < count:
-                LOG.error(_LE("Failed to allocate PCI devices for instance."
-                          " Unassigning devices back to pools."
-                          " This should not happen, since the scheduler"
-                          " should have accurate information, and allocation"
-                          " during claims is controlled via a hold"
-                          " on the compute node semaphore"))
+                LOG.error("Failed to allocate PCI devices for instance. "
+                          "Unassigning devices back to pools. "
+                          "This should not happen, since the scheduler "
+                          "should have accurate information, and allocation "
+                          "during claims is controlled via a hold "
+                          "on the compute node semaphore.")
                 for d in range(len(alloc_devices)):
                     self.add_device(alloc_devices.pop())
                 return None
@@ -210,26 +220,80 @@ class PciDeviceStats(object):
         return [pool for pool in pools
                 if utils.pci_device_prop_match(pool, request_specs)]
 
-    @staticmethod
-    def _filter_pools_for_numa_cells(pools, numa_cells):
-        # Some systems don't report numa node info for pci devices, in
-        # that case None is reported in pci_device.numa_node, by adding None
-        # to numa_cells we allow assigning those devices to instances with
-        # numa topology
-        numa_cells = [None] + [cell.id for cell in numa_cells]
-        # filter out pools which numa_node is not included in numa_cells
-        return [pool for pool in pools if any(utils.pci_device_prop_match(
-                                pool, [{'numa_node': cell}])
-                                              for cell in numa_cells)]
+    @classmethod
+    def _filter_pools_for_numa_cells(cls, pools, numa_cells, numa_policy,
+            requested_count):
+        """Filter out pools with the wrong NUMA affinity, if required.
 
-    def _filter_non_requested_pfs(self, request, matching_pools):
+        Exclude pools that do not have *suitable* PCI NUMA affinity.
+        ``numa_policy`` determines what *suitable* means, being one of
+        PREFERRED (nice-to-have), LEGACY (must-have-if-available) and REQUIRED
+        (must-have). We iterate through the various policies in order of
+        strictness. This means that even if we only *prefer* PCI-NUMA affinity,
+        we will still attempt to provide it if possible.
+
+        :param pools: A list of PCI device pool dicts
+        :param numa_cells: A list of InstanceNUMACell objects whose ``id``
+            corresponds to the ``id`` of host NUMACells.
+        :param numa_policy: The PCI NUMA affinity policy to apply.
+        :param requested_count: The number of PCI devices requested.
+        :returns: A list of pools that can, together, provide at least
+            ``requested_count`` PCI devices with the level of NUMA affinity
+            required by ``numa_policy``, else all pools that can satisfy this
+            policy even if it's not enough.
+        """
+        # NOTE(stephenfin): We may wish to change the default policy at a later
+        # date
+        requested_policy = numa_policy or fields.PCINUMAAffinityPolicy.LEGACY
+        numa_cell_ids = [cell.id for cell in numa_cells]
+
+        # filter out pools which numa_node is not included in numa_cell_ids
+        filtered_pools = [
+            pool for pool in pools if any(utils.pci_device_prop_match(
+                pool, [{'numa_node': cell}]) for cell in numa_cell_ids)]
+
+        # we can't apply a less strict policy than the one requested, so we
+        # need to return if we've demanded a NUMA affinity of REQUIRED.
+        # However, NUMA affinity is a good thing. If we can get enough devices
+        # with the stricter policy then we will use them.
+        if requested_policy == fields.PCINUMAAffinityPolicy.REQUIRED or sum(
+                pool['count'] for pool in filtered_pools) >= requested_count:
+            return filtered_pools
+
+        # some systems don't report NUMA node info for PCI devices, in which
+        # case None is reported in 'pci_device.numa_node'. The LEGACY policy
+        # allows us to use these devices so we include None in the list of
+        # suitable NUMA cells.
+        numa_cell_ids.append(None)
+
+        # filter out pools which numa_node is not included in numa_cell_ids
+        filtered_pools = [
+            pool for pool in pools if any(utils.pci_device_prop_match(
+                pool, [{'numa_node': cell}]) for cell in numa_cell_ids)]
+
+        # once again, we can't apply a less strict policy than the one
+        # requested, so we need to return if we've demanded a NUMA affinity of
+        # LEGACY. Similarly, we will also return if we have enough devices to
+        # satisfy this somewhat strict policy.
+        if requested_policy == fields.PCINUMAAffinityPolicy.LEGACY or sum(
+                pool['count'] for pool in filtered_pools) >= requested_count:
+            return filtered_pools
+
+        # if we've got here, we're using the PREFERRED policy and weren't able
+        # to provide anything with stricter affinity. Use whatever devices you
+        # can, folks.
+        return sorted(
+            pools, key=lambda pool: pool.get('numa_node') not in numa_cell_ids)
+
+    @classmethod
+    def _filter_non_requested_pfs(cls, pools, request):
         # Remove SRIOV_PFs from pools, unless it has been explicitly requested
-        # This is especially needed in cases where PFs and VFs has the same
+        # This is especially needed in cases where PFs and VFs have the same
         # product_id.
         if all(spec.get('dev_type') != fields.PciDeviceType.SRIOV_PF for
                spec in request.spec):
-            matching_pools = self._filter_pools_for_pfs(matching_pools)
-        return matching_pools
+            pools = cls._filter_pools_for_pfs(pools)
+        return pools
 
     @staticmethod
     def _filter_pools_for_pfs(pools):
@@ -237,16 +301,48 @@ class PciDeviceStats(object):
                 if not pool.get('dev_type') == fields.PciDeviceType.SRIOV_PF]
 
     def _apply_request(self, pools, request, numa_cells=None):
+        """Apply a PCI request.
+
+        Apply a PCI request against a given set of PCI device pools, which are
+        collections of devices with similar traits.
+
+        If ``numa_cells`` is provided then NUMA locality may be taken into
+        account, depending on the value of ``request.numa_policy``.
+
+        :param pools: A list of PCI device pool dicts
+        :param request: An InstancePCIRequest object describing the type,
+            quantity and required NUMA affinity of device(s) we want..
+        :param numa_cells: A list of InstanceNUMACell objects whose ``id``
+            corresponds to the ``id`` of host NUMACells.
+        :returns: True if the request was applied against the provided pools
+            successfully, else False.
+        """
         # NOTE(vladikr): This code maybe open to race conditions.
         # Two concurrent requests may succeed when called support_requests
         # because this method does not remove related devices from the pools
         count = request.count
+
+        # Firstly, let's exclude all devices that don't match our spec (e.g.
+        # they've got different PCI IDs or something)
         matching_pools = self._filter_pools_for_spec(pools, request.spec)
+
+        # Next, let's exclude all devices that aren't on the correct NUMA node
+        # *assuming* we have devices and care about that, as determined by
+        # policy
         if numa_cells:
+            numa_policy = None
+            if 'numa_policy' in request:
+                numa_policy = request.numa_policy
+
             matching_pools = self._filter_pools_for_numa_cells(matching_pools,
-                                                          numa_cells)
-        matching_pools = self._filter_non_requested_pfs(request,
-                                                        matching_pools)
+                numa_cells, numa_policy, count)
+
+        # Finally, if we're not requesting PFs then we should not use these.
+        # Exclude them.
+        matching_pools = self._filter_non_requested_pfs(matching_pools,
+                                                        request)
+
+        # Do we still have any devices left?
         if sum([pool['count'] for pool in matching_pools]) < count:
             return False
         else:
@@ -257,37 +353,52 @@ class PciDeviceStats(object):
         return True
 
     def support_requests(self, requests, numa_cells=None):
-        """Check if the pci requests can be met.
+        """Determine if the PCI requests can be met.
 
-        Scheduler checks compute node's PCI stats to decide if an
-        instance can be scheduled into the node. Support does not
-        mean real allocation.
-        If numa_cells is provided then only devices contained in
-        those nodes are considered.
+        Determine, based on a compute node's PCI stats, if an instance can be
+        scheduled on the node. **Support does not mean real allocation**.
+
+        If ``numa_cells`` is provided then NUMA locality may be taken into
+        account, depending on the value of ``numa_policy``.
+
+        :param requests: A list of InstancePCIRequest object describing the
+            types, quantities and required NUMA affinities of devices we want.
+        :type requests: nova.objects.InstancePCIRequests
+        :param numa_cells: A list of InstanceNUMACell objects whose ``id``
+            corresponds to the ``id`` of host NUMACells, or None.
+        :returns: Whether this compute node can satisfy the given request.
         """
         # note (yjiang5): this function has high possibility to fail,
         # so no exception should be triggered for performance reason.
         pools = copy.deepcopy(self.pools)
-        return all([self._apply_request(pools, r, numa_cells)
-                        for r in requests])
+        return all(self._apply_request(pools, r, numa_cells) for r in requests)
 
     def apply_requests(self, requests, numa_cells=None):
         """Apply PCI requests to the PCI stats.
 
         This is used in multiple instance creation, when the scheduler has to
         maintain how the resources are consumed by the instances.
-        If numa_cells is provided then only devices contained in
-        those nodes are considered.
+
+        If ``numa_cells`` is provided then NUMA locality may be taken into
+        account, depending on the value of ``numa_policy``.
+
+        :param requests: A list of InstancePCIRequest object describing the
+            types, quantities and required NUMA affinities of devices we want.
+        :type requests: nova.objects.InstancePCIRequests
+        :param numa_cells: A list of InstanceNUMACell objects whose ``id``
+            corresponds to the ``id`` of host NUMACells, or None.
+        :raises: exception.PciDeviceRequestFailed if this compute node cannot
+            satisfy the given request.
         """
-        if not all([self._apply_request(self.pools, r, numa_cells)
-                                            for r in requests]):
+        if not all(self._apply_request(self.pools, r, numa_cells)
+                   for r in requests):
             raise exception.PciDeviceRequestFailed(requests=requests)
 
     def __iter__(self):
         # 'devices' shouldn't be part of stats
         pools = []
         for pool in self.pools:
-            tmp = {k: v for k, v in six.iteritems(pool) if k != 'devices'}
+            tmp = {k: v for k, v in pool.items() if k != 'devices'}
             pools.append(tmp)
         return iter(pools)
 
@@ -298,10 +409,6 @@ class PciDeviceStats(object):
     def __eq__(self, other):
         return self.pools == other.pools
 
-    if six.PY2:
-        def __ne__(self, other):
-            return not (self == other)
-
     def to_device_pools_obj(self):
         """Return the contents of the pools as a PciDevicePoolList object."""
         stats = [x for x in self]
